from __future__ import unicode_literals

import dbus
import types
import pprint

from exceptions import BTSignalNameNotRecognisedException


def translate_to_dbus_type(typeof, value):
    """
    Helper function to map values from their native Python types
    to Dbus types.

    :param type typeof: Target for type conversion e.g., 'dbus.Dictionary'
    :param value: Value to assign using type 'typeof'
    :return: 'value' converted to type 'typeof'
    :rtype: typeof
    """
    if ((isinstance(value, types.UnicodeType) or
         isinstance(value, str)) and typeof is not dbus.String):
        # FIXME: This is potentially dangerous since it evaluates
        # a string in-situ
        return typeof(eval(value))
    else:
        return typeof(value)


class Signal():
    """
    Encapsulation of user callback wrapper for signals
    fired by dbus.  This allows us to prepend the signal
    name and the user callback argument.

    :param str signal: Signal name
    :param func user_callback: User-defined callback function to
        call when the signal triggers
    :param user_arg: User-defined callback argument to be passed
        as callback function
    """
    def __init__(self, signal, user_callback, user_arg):
        self.signal = signal
        self.user_callback = user_callback
        self.user_arg = user_arg

    def signal_handler(self, *args):
        """
        Method to call in order to invoke the user callback.

        :param args: list of signal-dependent arguments
        :return:
        """
        self.user_callback(self.signal, self.user_arg, *args)


class BTSimpleInterface:
    """
    Wrapper around dbus to encapsulated a BT simple interface
    entry point (i.e., has no signals or properties).

    :param str path: Object path pertaining to the interface to open
                     e.g., '/org/bluez/985/hci0'
    :param str addr: dbus address of the interface instance to open
                     e.g., 'org.bluez.Adapter'

    .. note:: This class should always be sub-classed with a concrete
        implementation of a bluez interface which has no signals or
        properties.
    """
    def __init__(self, path, addr):
        self._dbus_addr = addr
        self._bus = dbus.SystemBus()
        self._object = self._bus.get_object('org.bluez', path)
        self._interface = dbus.Interface(self._object, addr)


# This class is not intended to be instantiated directly and should be
# sub-classed with a concrete implementation for an interface
class BTInterface(BTSimpleInterface):
    """
    Wrapper around DBus to encapsulated a BT interface
    entry point e.g., an adapter, a device, etc.

    :param str path: Object path pertaining to the interface to open
                     e.g., '/org/bluez/985/hci0'
    :param str addr: dbus address of the interface instance to open
                     e.g., 'org.bluez.Adapter'

    .. note:: This class should always be sub-classed with a concrete
        implementation of a bluez interface which has both signals
        and properties.
    """

    SIGNAL_PROPERTY_CHANGED = 'PropertyChanged'
    """
    :signal PropertyChanged(sig_name, user_arg, prop_name, prop_value):
        Signal notifying when a property has changed.
    """

    def __init__(self, path, addr):
        BTSimpleInterface.__init__(self, path, addr)
        self._signals = {}
        self._signal_names = []
        self._properties = self._interface.GetProperties().keys()
        self._register_signal_name(BTInterface.SIGNAL_PROPERTY_CHANGED)

    def _register_signal_name(self, name):
        """
        Helper function to register allowed signals on this
        instance.  Need only be called once per signal name and must be done
        for each signal that may be used via :py:meth:`add_signal_receiver`

        :param str name: Signal name to register e.g.,
            :py:attr:`SIGNAL_PROPERTY_CHANGED`
        :return:
        """
        self._signal_names.append(name)

    def add_signal_receiver(self, callback_fn, signal, user_arg):
        """
        Add a signal receiver callback with user argument

        See also :py:meth:`remove_signal_receiver`,
        :py:exc:`.BTSignalNameNotRecognisedException`

        :param func callback_fn: User-defined callback function to call when
            signal triggers
        :param str signal: Signal name e.g.,
            :py:attr:`.BTInterface.SIGNAL_PROPERTY_CHANGED`
        :param user_arg: User-defined callback argument to be passed with
            callback function
        :return:
        :raises BTSignalNameNotRecognisedException: if the signal name is
            not registered
        """
        if (signal in self._signal_names):
            s = Signal(signal, callback_fn, user_arg)
            self._signals[signal] = s
            self._bus.add_signal_receiver(s.signal_handler,
                                          signal,
                                          dbus_interface=self._dbus_addr)
        else:
            raise BTSignalNameNotRecognisedException

    def remove_signal_receiver(self, signal):
        """
        Remove an installed signal receiver by signal name.

        See also :py:meth:`add_signal_receiver`
        :py:exc:`exceptions.BTSignalNameNotRecognisedException`

        :param str signal: Signal name to uninstall
            e.g., :py:attr:`SIGNAL_PROPERTY_CHANGED`
        :return:
        :raises BTSignalNameNotRecognisedException: if the signal name is
            not registered
        """
        if (signal in self._signal_names):
            s = self._signals.get(signal)
            if (s):
                self._bus.remove_signal_receiver(s.signal_handler,
                                                 signal,
                                                 dbus_interface=self._dbus_addr)  # noqa
                self._signals.pop(signal)
        else:
            raise BTSignalNameNotRecognisedException

    def get_property(self, name=None):
        """
        Helper to get a property value by name or all
        properties as a dictionary.

        See also :py:meth:`set_property`

        :param str name: defaults to None which means all properties
            in the object's dictionary are returned as a dict.
            Otherwise, the property name key is used and its value
            is returned.
        :return: Property value by property key, or a dictionary of
            all properties
        :raises KeyError: if the property key is not found in the
            object's dictionary
        :raises dbus.Exception: org.bluez.Error.DoesNotExist
        :raises dbus.Exception: org.bluez.Error.InvalidArguments
        """
        if (name):
            return self._interface.GetProperties()[name]
        else:
            return self._interface.GetProperties()

    def set_property(self, name, value):
        """
        Helper to set a property value by name, translating to correct
        dbus type

        See also :py:meth:`get_property`

        :param str name: The property name in the object's dictionary
            whose value shall be set.
        :param value: Properties new value to be assigned.
        :return:
        :raises KeyError: if the property key is not found in the
            object's dictionary
        :raises dbus.Exception: org.bluez.Error.DoesNotExist
        :raises dbus.Exception: org.bluez.Error.InvalidArguments
        """
        typeof = type(self.get_property(name))
        self._interface.SetProperty(name,
                                    translate_to_dbus_type(typeof, value))

    def __getattr__(self, name):
        """Override default getattr behaviours to allow DBus object
        properties to be exposed in the class for getting"""
        if name in self.__dict__:
            return self.__dict__[name]
        elif '_properties' in self.__dict__ and name in self._properties:
            return self.get_property(name)

    def __setattr__(self, name, value):
        """Override default setattr behaviours to allow DBus object
        properties to be exposed in the class for setting"""
        if '_properties' in self.__dict__ and name not in self.__dict__:
            self.set_property(name, value)
        else:
            self.__dict__[name] = value

    def __repr__(self):
        """Stringify the Dbus interface properties as raw"""
        return self.__str__()

    def __str__(self):
        """Stringify the Dbus interface properties in a nice format"""
        return pprint.pformat(self._interface.GetProperties())



from __future__ import unicode_literals

from interface import BTInterface


class BTManager(BTInterface):
    """
    Wrapper around dbus to encapsulate the org.bluez.manager interface
    which notionally is used to manage available bluetooth adapters.

    :Properties:

    * **Adapters(list{str}) [readonly]**: List of adapter object paths.

    See also :py:class:`.BTAdapter`
    """

    SIGNAL_ADAPTER_ADDED = 'AdapterAdded'
    """
    :signal AdapterAdded(signal_name, user_arg, object_path):
        Signal notifying when an adapter is added.
    """
    SIGNAL_ADAPTER_REMOVED = 'AdapterRemoved'
    """
    :signal AdapterRemoved(signal_name, user_arg, object_path):
        Signal notifying when an adapter is removed.
        .. note: In case all adapters are removed this signal will not
        be emitted. The AdapterRemoved signal has to be used to
        detect that no default adapter is selected or available
        anymore.
    """
    SIGNAL_DEFAULT_ADAPTER_CHANGED = 'DefaultAdapterChanged'
    """
    :signal DefaultAdapterChanged(signal_name, user_arg, object_path):
        Signal notifying when the default adapter has been changed.
    """

    def __init__(self):
        BTInterface.__init__(self, '/', 'org.bluez.Manager')
        self._register_signal_name(BTManager.SIGNAL_ADAPTER_ADDED)
        self._register_signal_name(BTManager.SIGNAL_ADAPTER_REMOVED)
        self._register_signal_name(BTManager.SIGNAL_DEFAULT_ADAPTER_CHANGED)

    def default_adapter(self):
        """
        Obtain the default BT adapter object path.

        :return: Object path of default adapter
        :rtype: str
        :raises dbus.Exception: org.bluez.Error.InvalidArguments
        :raises dbus.Exception: org.bluez.Error.NoSuchAdapter
        """
        return self._interface.DefaultAdapter()

    def find_adapter(self, pattern):
        """
        Returns object path for the specified adapter.

        :param str pattern:  Valid patterns are "hci0" or "00:11:22:33:44:55".
        :return: Object path of adapter
        :rtype: str
        :raises dbus.Exception: org.bluez.Error.InvalidArguments
        :raises dbus.Exception: org.bluez.Error.NoSuchAdapter
        """
        return self._interface.FindAdapter(pattern)

    def list_adapters(self):
        """
        Returns list of adapter object paths under /org/bluez

        :return: List of object paths or each adapter attached
        :rtype: list
        :raises dbus.Exception: org.bluez.Error.InvalidArguments
        :raises dbus.Exception: org.bluez.Error.Failed
        :raises dbus.Exception: org.bluez.Error.OutOfMemory
        """
        return self._interface.ListAdapters()



from __future__ import unicode_literals


class BTCoD:
    """
    Bluetooth class of device decoder for providing
    a human readable form of the device class value.

    The class of device is a 24-bit number allowing different
    types of bluetooth devices to be described consistently
    and succinctly using a hierarchical scheme.

    * **Major Service**: The major services that the
        device supports.  Several values can be OR'd together
        to form a device supporting multiple major services
        e.g., telephony, networking and audio could be
        the combined service set for a device.
    * **Major Device**: The form-factor or major type of device
        which may only take a single value e.g., Laptop,
        Toy, Health Device, Phone, etc.
    * **Minor Device**: The minor function of the device whose
        range of values depends on the major function e.g.,
        Car audio, SmartPhone, etc.

    :param integer cod: 24-bit integer value representing the
        class of device.

    .. note:: The class of device may be obtained from both
        the :py:class:`.BTAdapter` and :py:class:`.BTDevice`
        classes using the `Class` attribute.
    """

    _MAJOR_SERVICE_POS = 13
    _MAJOR_SERVICE_MASK = 0xFFE000
    _MAJOR_SERVICE_CLASS = {
        0x002000: 'Limited Discoverable Mode [Ref #1]',
        0x004000: '(reserved)',
        0x008000: '(reserved)',
        0x010000: 'Positioning (Location identification)',
        0x020000: 'Networking (LAN, Ad hoc, ...)',
        0x040000: 'Rendering (Printing, Speakers, ...)',
        0x080000: 'Capturing (Scanner, Microphone, ...)',
        0x100000: 'Object Transfer (v-Inbox, v-Folder, ...)',
        0x200000: 'Audio (Speaker, Microphone, Headset service, ...)',
        0x400000: 'Telephony (Cordless telephony, Modem, Headset service, ...)',  # noqa
        0x800000: 'Information (WEB-server, WAP-server, ...)',
    }

    _MAJOR_DEVICE_POS = 8
    _MAJOR_DEVICE_MASK = 0x001F00
    _MAJOR_DEVICE_CLASS = {
        0x0000: 'Miscellaneous [Ref #2]',
        0x0100: 'Computer (desktop, notebook, PDA, organizer, ... )',
        0x0200: 'Phone (cellular, cordless, pay phone, modem, ...)',
        0x0300: 'LAN /Network Access point',
        0x0400: 'Audio/Video (headset, speaker, stereo, video display, VCR, ...',  # noqa
        0x0500: 'Peripheral (mouse, joystick, keyboard, ... ',
        0x0600: 'Imaging (printer, scanner, camera, display, ...)',
        0x0700: 'Wearable',
        0x0800: 'Toy',
        0x0900: 'Health',
        0x1F00: 'Uncategorized: device code not specified'
    }

    _MINOR_DEVICE_CLASS = {
        0x0100: [{'mask': 0x1C, 'pos': 2,
                  0x00: 'Uncategorized, code for device not assigned',
                  0x04: 'Desktop workstation',
                  0x08: 'Server-class computer',
                  0x0C: 'Laptop',
                  0x10: 'Handheld PC/PDA (clamshell)',
                  0x14: 'Palm-size PC/PDA',
                  0x18: 'Wearable computer (watch size)',
                  0x1C: 'Tablet'}],
        0x0200: [{'mask': 0x1C, 'pos': 2,
                  0x00: 'Uncategorized, code for device not assigned',
                  0x04: 'Cellular',
                  0x08: 'Cordless',
                  0x0C: 'Smartphone',
                  0x10: 'Wired modem or voice gateway',
                  0x14: 'Common ISDN access'}],
        0x0300: [{'mask': 0xE0, 'pos': 5,
                  0x00: 'Fully available',
                  0x20: '1% to 17% utilized',
                  0x40: '17% to 33% utilized',
                  0x60: '33% to 50% utilized',
                  0x80: '50% to 67% utilized',
                  0xA0: '67% to 83% utilized',
                  0xC0: '83% to 99% utilized',
                  0xE0: 'No service available'}],
        0x0400: [{'mask': 0x7C, 'pos': 2,
                  0x00: 'Uncategorized, code not assigned',
                  0x04: 'Wearable Headset Device',
                  0x08: 'Hands-free Device',
                  0x0C: '(Reserved)',
                  0x10: 'Microphone',
                  0x14: 'Loudspeaker',
                  0x18: 'Headphones',
                  0x1C: 'Portable Audio',
                  0x20: 'Car audio',
                  0x24: 'Set-top box',
                  0x28: 'HiFi Audio Device',
                  0x2C: 'VCR',
                  0x30: 'Video Camera',
                  0x34: 'Camcorder',
                  0x38: 'Video Monitor',
                  0x3C: 'Video Display and Loudspeaker',
                  0x40: 'Video Conferencing',
                  0x44: '(Reserved)',
                  0x48: 'Gaming/Toy'}],
        0x0500: [{'mask': 0xC0, 'pos': 6,
                  0x00: 'Not Keyboard / Not Pointing Device',
                  0x40: 'Keyboard',
                  0x80: 'Pointing device',
                  0xC0: 'Combo keyboard/pointing device'},
                 {'mask': 0x3C, 'pos': 2,
                  0x00: 'Uncategorized device',
                  0x04: 'Joystick',
                  0x08: 'Gamepad',
                  0x0C: 'Remote control',
                  0x10: 'Sensing device',
                  0x14: 'Digitizer tablet',
                  0x18: 'Card Reader (e.g. SIM Card Reader)',
                  0x1C: 'Digital Pen',
                  0x20: 'Handheld scanner for bar-codes, RFID, etc.',
                  0x24: 'Handheld gestural input device (e.g., "wand" form factor)'}],  # noqa
        0x0600: [{'mask': 0xF0, 'pos': 4,
                  0x10: 'Display',
                  0x20: 'Camera',
                  0x40: 'Scanner',
                  0x80: 'Printer'}],
        0x0700: [{'mask': 0x1C, 'pos': 2,
                  0x04: 'Wristwatch',
                  0x08: 'Pager',
                  0x0C: 'Jacket',
                  0x10: 'Helmet',
                  0x14: 'Glasses'}],
        0x0800: [{'mask': 0x1C, 'pos': 2,
                  0x04: 'Robot',
                  0x08: 'Vehicle',
                  0x0C: 'Doll / Action figure',
                  0x10: 'Controller',
                  0x14: 'Game'}],
        0x0900: [{'mask': 0x3C, 'pos': 2,
                  0x00: 'Undefined',
                  0x04: 'Blood Pressure Monitor',
                  0x08: 'Thermometer',
                  0x0C: 'Weighing Scale',
                  0x10: 'Glucose Meter',
                  0x14: 'Pulse Oximeter',
                  0x18: 'Heart/Pulse Rate Monitor',
                  0x1C: 'Health Data Display',
                  0x20: 'Step Counter',
                  0x24: 'Body Composition Analyzer',
                  0x28: 'Peak Flow Monitor',
                  0x2C: 'Medication Monitor',
                  0x30: 'Knee Prosthesis',
                  0x34: 'Ankle Prosthesis',
                  0x38: 'Generic Health Manager',
                  0x3C: 'Personal Mobility Device'}]
    }

    def __init__(self, cod):
        self.cod = cod

    @property
    def major_service_class(self):
        """
        Return the major service class property decoded e.g.,
        Audio, Telephony, etc
        """
        major_service = []
        for i in BTCoD._MAJOR_SERVICE_CLASS.keys():
            if (self.cod & i):
                major_service.append(BTCoD._MAJOR_SERVICE_CLASS[i])
        return major_service

    @property
    def major_device_class(self):
        """
        Return the major device class property decoded e.g.,
        Computer, Audio/Video, Toy, etc.
        """
        return BTCoD._MAJOR_DEVICE_CLASS.get(self.cod &
                                             BTCoD._MAJOR_DEVICE_MASK,
                                             'Unknown')

    @property
    def minor_device_class(self):
        """
        Return the minor device class property decoded e.g.,
        Scanner, Printer, Loudspeaker, Camera, etc.
        """
        minor_device = []
        minor_lookup = BTCoD._MINOR_DEVICE_CLASS.get(self.cod &
                                                     BTCoD._MAJOR_DEVICE_MASK,
                                                     [])
        for i in minor_lookup:
            minor_value = self.cod & i.get('mask')
            minor_device.append(i.get(minor_value, 'Unknown'))
        return minor_device

    def __str__(self):
        """Stringify all elements of the class of device"""
        return '<cod:' + str(hex(self.cod)) + ' Major Service:' + str(self.major_service_class) + \
            ' Major Device:' + \
            self.major_device_class + ' Minor Device:' + \
            str(self.minor_device_class) + '>'

    def __repr__(self):
        return self.__str__()



from __future__ import unicode_literals

import dbus


class BTSignalNameNotRecognisedException:
    """
    Exception raised for when a signal name is not recognized.
    Check the originating class for a list of supported signal names
    """
    pass


class BTDeviceNotSpecifiedException:
    """
    Exception raised for when a device is not specified
    """
    pass


class BTRejectedException(dbus.DBusException):
    """
    Exception raised to notify that the user rejected
    a pairing attempt.
    """
    _dbus_error_name = "org.bluez.Error.Rejected"


class BTInvalidConfiguration(dbus.DBusException):
    """
    Exception raised to denote an invalid configuration
    parameter
    """
    _dbus_error_name = "org.bluez.Error.InvalidConfiguration"


class BTIncompatibleTransportAccessType(dbus.DBusException):
    """
    Exception raised when attempting to access a media
    transport file descriptor without correct access
    permissions.
    """
    _dbus_error_name = "org.bluez.Error.InvalidConfiguration"


class BTUUIDNotSpecifiedException:
    """
    Exception raised when creating a UUID without providing
    a valid UUID number
    """
    pass



from __future__ import unicode_literals

import dbus.service

from exceptions import BTRejectedException


class BTAgent(dbus.service.Object):
    """
    Simple BT device pairing agent.

    A bluetooth pairing agent is responsible for admitting
    new devices onto the system and registering them
    for subsequent use.  The procedure is abstracted away
    into a service object that implements a few procedures
    depending on the type of pairing scheme in use.

    Different pairing schemes are required since some
    devices do not have displays or keypads allowing
    them to enter passkeys or confirm passkeys.

    The SSP (Secure Simple Pairing) schemes that can
    be supported using this agent are:

    * `Just works`: No user interaction required.
    * `Numeric comparison`: A pass key is displayed which
        the receiving device must confirm with a
        binary yes/no response.
    * `Passkey entry`: A pass key must be entered manually
        which the receiving device must confirm by also
        entering the same pass key.

    The pairing scheme selected will be dependent on
    the `capability` of the agent which is defined
    when the agent is registered via the
    :py:meth:`.BTAdapter.register_agent` method.

    See also: :py:class:`.BTAdapter`
    """

    NOTIFY_ON_RELEASE = 'Release'
    NOTIFY_ON_AUTHORIZE = 'Authorize'
    NOTIFY_ON_REQUEST_PIN_CODE = 'RequestPinCode'
    NOTIFY_ON_REQUEST_PASS_KEY = 'RequestPasskey'
    NOTIFY_ON_DISPLAY_PASS_KEY = 'DisplayPasskey'
    NOTIFY_ON_REQUEST_CONFIRMATION = 'RequestConfirmation'
    NOTIFY_ON_CONFIRM_MODE_CHANGE = 'ConfirmModeChange'
    NOTIFY_ON_CANCEL = 'Cancel'

    def __init__(self,
                 path='/test/agent',
                 auto_authorize_connections=True,
                 default_pin_code='0000',
                 default_pass_key=0,   # Range: 0-999999
                 cb_notify_on_release=None,
                 cb_notify_on_authorize=None,
                 cb_notify_on_request_pin_code=None,
                 cb_notify_on_request_pass_key=None,
                 cb_notify_on_display_pass_key=None,
                 cb_notify_on_request_confirmation=None,
                 cb_notify_on_confirm_mode_change=None,
                 cb_notify_on_cancel=None):

        self.auto_authorize_connections = auto_authorize_connections
        self.default_pin_code = default_pin_code
        self.default_pass_key = default_pass_key
        self.cb_notify_on_release = cb_notify_on_release
        self.cb_notify_on_authorize = cb_notify_on_authorize
        self.cb_notify_on_request_pin_code = cb_notify_on_request_pin_code
        self.cb_notify_on_request_pass_key = cb_notify_on_request_pass_key
        self.cb_notify_on_display_pass_key = cb_notify_on_display_pass_key
        self.cb_notify_on_request_confirmation = \
            cb_notify_on_request_confirmation
        self.cb_notify_on_confirm_mode_change = \
            cb_notify_on_confirm_mode_change
        self.cb_notify_on_cancel = cb_notify_on_cancel
        bus = dbus.SystemBus()
        super(BTAgent, self).__init__(bus, path)

    # Service object entry points defined below here
    @dbus.service.method("org.bluez.Agent", in_signature="", out_signature="")
    def Release(self):
        if (self.cb_notify_on_release):
            self.cb_notify_on_release(BTAgent.NOTIFY_ON_RELEASE)

    @dbus.service.method("org.bluez.Agent", in_signature="os",
                         out_signature="")
    def Authorize(self, device, uuid):
        if (self.cb_notify_on_authorize):
            if (not self.cb_notify_on_authorize(BTAgent.NOTIFY_ON_AUTHORIZE,
                                                device,
                                                uuid)):
                raise BTRejectedException('Connection not authorized by user')
        elif (not self.auto_authorize_connections):
            raise BTRejectedException('Auto authorize is off')

    @dbus.service.method("org.bluez.Agent", in_signature="o",
                         out_signature="s")
    def RequestPinCode(self, device):
        if (self.cb_notify_on_request_pin_code):
            pin_code = self.cb_notify_on_request_pin_code(BTAgent.NOTIFY_ON_REQUEST_PIN_CODE,  # noqa
                                                          device)
            if (pin_code is None):
                raise BTRejectedException('User did not provide PIN code')
        elif (self.default_pin_code is None):
            raise BTRejectedException('No default PIN code set')
        else:
            pin_code = self.default_pin_code
        return dbus.String(pin_code)

    @dbus.service.method("org.bluez.Agent", in_signature="o",
                         out_signature="s")
    def RequestPasskey(self, device):
        if (self.cb_notify_on_request_pass_key):
            pass_key = self.cb_notify_on_request_pass_key(BTAgent.NOTIFY_ON_REQUEST_PASS_KEY,  # noqa
                                                          device)
            if (pass_key is None):
                raise BTRejectedException('User did not provide pass key')
        elif (self.default_pass_key is None):
            raise BTRejectedException('No default pass key set')
        else:
            pass_key = self.default_pass_key
        return dbus.UInt32(pass_key)

    @dbus.service.method("org.bluez.Agent", in_signature="ou",
                         out_signature="")
    def DisplayPasskey(self, device, pass_key):
        if (self.cb_notify_on_display_pass_key):
            self.cb_notify_on_display_pass_key(BTAgent.NOTIFY_ON_DISPLAY_PASS_KEY,  # noqa
                                               device, pass_key)

    @dbus.service.method("org.bluez.Agent", in_signature="ou",
                         out_signature="")
    def RequestConfirmation(self, device, pass_key):
        if (self.cb_notify_on_request_confirmation):
            if (not self.cb_notify_on_request_confirmation(BTAgent.NOTIFY_ON_REQUEST_CONFIRMATION,  # noqa
                                                           device, pass_key)):
                raise \
                    BTRejectedException('User confirmation of pass key negative')  # noqa

    @dbus.service.method("org.bluez.Agent", in_signature="s", out_signature="")
    def ConfirmModeChange(self, mode):
        if (self.cb_notify_on_confirm_mode_change):
            if (not self.cb_notify_on_confirm_mode_change(BTAgent.NOTIFY_ON_CONFIRM_MODE_CHANGE,  # noqa
                                                          mode)):
                raise \
                    BTRejectedException('User mode change confirmation negative')  # noqa

    @dbus.service.method("org.bluez.Agent", in_signature="", out_signature="")
    def Cancel(self):
        if (self.cb_notify_on_cancel):
            self.cb_notify_on_cancel(BTAgent.NOTIFY_ON_CANCEL)



from __future__ import unicode_literals

import unittest

import bt_manager


class UUIDTest(unittest.TestCase):

    def test_uuid(self):
        name = 'UUID Name'
        desc = 'UUID Description'
        uuid = '12345678-xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx'
        obj = bt_manager.BTUUID(uuid=uuid, name=name, desc=desc)
        self.assertEqual(obj.uuid, uuid.upper())
        self.assertEqual(obj.uuid16, uuid[4:8])
        self.assertEqual(obj.uuid32, uuid[0:8])
        self.assertEqual(obj.name, name)
        self.assertEqual(obj.desc, desc)

    def test_uuid16(self):
        name = 'UUID Name'
        desc = 'UUID Description'
        uuid = '1234'
        obj = bt_manager.BTUUID16(uuid=uuid, name=name, desc=desc)
        self.assertEqual(obj.uuid16, uuid)
        self.assertEqual(obj.name, name)
        self.assertEqual(obj.desc, desc)

    def test_uuid32(self):
        name = 'UUID Name'
        desc = 'UUID Description'
        uuid = '12345678'
        obj = bt_manager.BTUUID32(uuid=uuid, name=name, desc=desc)
        self.assertEqual(obj.uuid32, uuid)
        self.assertEqual(obj.name, name)
        self.assertEqual(obj.desc, desc)

    def test_base_uuid(self):
        uuid = '00000000-0000-1000-8000-00805F9B34FB'
        self.assertEqual(bt_manager.BASE_UUID.uuid, uuid)

    def test_UUIDNotSpecifiedException(self):
        try:
            caught = False
            bt_manager.BTUUID()
        except bt_manager.BTUUIDNotSpecifiedException:
            caught = True
        self.assertTrue(caught)



from __future__ import unicode_literals
from distutils.version import StrictVersion
import cffi
import os

__version__ = '0.1.0'

if StrictVersion(cffi.__version__) < StrictVersion('0.7'):
        raise RuntimeError(
            'bt_manager requires cffi >= 0.7, but found %s' % cffi.__version__)

ffi = cffi.FFI()
cwd = os.path.dirname(__file__)
header_file = os.path.join(cwd, 'rtpsbc.h')
with open(header_file) as fh:
    header = fh.read()
    ffi.cdef(header)
    fh.close()

from bt_manager.adapter import BTAdapter                 # noqa
from bt_manager.agent import BTAgent                     # noqa
from bt_manager.attributes import ATTRIBUTES             # noqa
from bt_manager.audio import BTAudio, BTAudioSource      # noqa
from bt_manager.audio import BTAudioSink, SBCAudioCodec  # noqa
from bt_manager.audio import SBCAudioSource, SBCAudioSink  # noqa
from bt_manager.cod import BTCoD                         # noqa
from bt_manager.codecs import *                          # noqa
from bt_manager.control import BTControl                 # noqa
from bt_manager.device import BTGenericDevice, BTDevice  # noqa
from bt_manager.discovery import BTDiscoveryInfo         # noqa
from bt_manager.exceptions import *                      # noqa
from bt_manager.headset import BTHeadset                 # noqa
from bt_manager.headset import BTHeadsetGateway          # noqa
from bt_manager.interface import BTSimpleInterface       # noqa
from bt_manager.interface import BTInterface             # noqa
from bt_manager.manager import BTManager                 # noqa
from bt_manager.media import BTMedia, BTMediaTransport   # noqa
from bt_manager.serviceuuids import SERVICES             # noqa
from bt_manager.uuid import BTUUID, BTUUID16, BTUUID32   # noqa
from bt_manager.uuid import BASE_UUID                    # noqa
from bt_manager.vendors import VENDORS                   # noqa



from __future__ import unicode_literals
from exceptions import BTUUIDNotSpecifiedException


_BASE_UUID = '00000000-0000-1000-8000-00805F9B34FB'


class BTUUID:
    """
    This class encapsulates a UUID (universally unique identifier)
    which is a 128-bit value.  It is represented as a hex string of
    the form `ZZZZYYYY-BBBB-BBBB-BBBB-VVVVVVVVVVVV` although
    typically UUIDs are referenced only by the 16-bit `YYYY` part
    since this is generally sufficient to uniquely identify
    the services/protocols that form part of the bluetooth
    standard.

    When creating a UUID as 16-bits only, the `YYYY` nibbles are
    set accordingly with the remaining nibbles set automatically
    using the base UUID which is
    `00000000-0000-1000-8000-00805F9B34FB`.

    When creating a UUID as 32-bits, the uppermost 8 nibbles shall
    be set instead.

    Otherwise a fully qualified 128-bit UUID is assumed to be
    provided.

    For ease of use, optionally, a human readable name and
    description can be provided with the UUID.  It is encouraged
    that this is done as UUIDs are otherwise hard to read.

    :param str uuid: Optional full 128-bit UUID string of the
        form `ZZZZYYYY-BBBB-BBBB-BBBB-VVVVVVVVVVVV`
    :param str uuid16: Optional 16-bit UUID string of the form
        `YYYY`.  Base UUID is used to populate unset bits.
    :param str uuid32: Optional 32-bit UUID string of the form
        `ZZZZYYYY`.    Base UUID is used to populate unset
        bits.
    :param str name: Optional string providing a unique name
        for the UUID.
    :param str desc: Optional string providing a description
        of the UUID.
    :raises BTUUIDNotSpecifiedException: if neither a uuid,
        uuid16 nor uuid32 is provided.
    """
    def __init__(self, uuid=None, uuid16=None,
                 uuid32=None, name=None, desc=None):
        self.name = name
        self.desc = desc
        if (uuid):
            self.uuid = uuid.upper()
        elif (uuid16):
            self.uuid = _BASE_UUID[0:4] + uuid16[0:4].upper() + _BASE_UUID[8:]
        elif (uuid32):
            self.uuid = uuid32[0:8].upper() + _BASE_UUID[8:]
        else:
            raise BTUUIDNotSpecifiedException

    @property
    def uuid16(self):
        """
        Returns the 16-bit part of the UUID string i.e., given
        a UUID of the form `ZZZZYYYY-BBBB-BBBB-BBBB-VVVVVVVVVVVV`
        it shall return `YYYY`
        """
        return self.uuid[4:8]

    @property
    def uuid32(self):
        """
        Returns the 32-bit part of the UUID string i.e., given
        a UUID of the form `ZZZZYYYY-BBBB-BBBB-BBBB-VVVVVVVVVVVV`
        it shall return `ZZZZYYYY`
        """
        return self.uuid[0:8]

    def __repr__(self):
        return '<uuid:' + self.uuid + ' name:' + \
            str(self.name) + ' desc:' + str(self.desc) + '>'


class BTUUID16(BTUUID):
    """
    Shortened-form UUID allowing only 16-bit UUID to be
    created.  Refer to :py:class:`BTUUID` for details.
    """
    def __init__(self, uuid, name, desc=None):
        BTUUID.__init__(self, uuid16=uuid, name=name, desc=desc)


class BTUUID32(BTUUID):
    """
    Shortened-form UUID allowing only 32-bit UUID to be
    created.  Refer to :py:class:`BTUUID` for details.
    """
    def __init__(self, uuid, name, desc=None):
        BTUUID.__init__(self, uuid32=uuid, name=name, desc=desc)


BASE_UUID = BTUUID(uuid=_BASE_UUID, name='BASE_UUID',
                   desc='Base Universally Unique Identifier')
"""
:data BASE_UUID: The base UUID which takes the value of
    `00000000-0000-1000-8000-00805F9B34FB`
"""



from __future__ import unicode_literals

__SupportedFeatureOnly = {
    '0311': 'SupportedFeatures',
}

__BIP = {
    '0311': 'SupportedFeatures',
    '0200': 'GoepL2capPsm',
    '0310': 'SupportedCapabilities',
    '0312': 'SupportedFunctions',
    '0313': 'TotalImagingDataCapacity',
}

__BPP = {
    '0350': 'Document Formats Supported',
    '0352': 'Character Repertoires Supported',
    '0354': 'XHTML-Print Image Formats Supported',
    '0356': 'Color Supported',
    '0358': '1284ID',
    '035A': 'Printer Name',
    '035C': 'Printer Location',
    '035E': 'Duplex Supported',
    '0360': 'Media Types Supported',
    '0362': 'MaxMediaWidth',
    '0364': 'MaxMediaLength',
    '0366': 'Enhanced Layout Supported',
    '0368': 'RUI Formats Supported',
    '0370': 'Reference Printing RUI Supported',
    '0372': 'Direct Printing RUI Supported',
    '0374': 'Reference Printing Top URL',
    '0376': 'Direct Printing Top URL',
    '0378': 'Printer Admin RUI Top URL',
    '037A': 'Device Name',
}

__Universal = {
    '0000': 'ServiceRecordHandle',
    '0001': 'ServiceClassIDList',
    '0002': 'ServiceRecordState',
    '0003': 'ServiceID',
    '0004': 'ProtocolDescriptorList',
    '0005': 'BrowseGroupList',
    '0006': 'LanguageBaseAttributeIDList',
    '0007': 'ServiceInfoTimeToLive',
    '0008': 'ServiceAvailability',
    '0009': 'BluetoothProfileDescriptorList',
    '000A': 'DocumentationURL',
    '000B': 'ClientExecutableURL',
    '000C': 'IconURL',
    '000D': 'AdditionalProtocolDescriptorLists',
}

__ServiceDiscoveryService = {
    '0200': 'VersionNumberList',
    '0201': 'ServiceDatabaseState',
}

__BrowseGroupDescriptorService = {
    '0200': 'GroupID',
}

__CordlessTelephony = {
    '0301': 'External Network',
}

__DeviceIdentification = {
    '0200': 'SpecificationID',
    '0201': 'VendorID',
    '0202': 'ProductID',
    '0203': 'Version',
    '0204': 'PrimaryRecord',
    '0205': 'VendorIDSource',
}

__Fax = {
    '0302': 'Fax Class 1 Support',
    '0303': 'Fax Class 2.0 Support',
    '0304': 'Fax Class 2 Support(vendor-specific class)',
    '0305': 'Audio Feedback Support',
}

__FTP = {
    '0200': 'GoepL2capPsm(FTP v1.2 and later)',
}

__GNSSService = {
    '0200': 'SupportedFeatures'
}

__HandsFreeProfile = {
    '0301': 'Network',
    '0311': 'SupportedFeatures',
}

__HardcopyReplacementProfile = {
    '0300': '1284ID',
    '0302': 'Device Name',
    '0304': 'Friendly Name',
    '0306': 'Device Location',
}

__HeadsetProfile = {
    '0302': 'Remote Audio Volume Control',
}

__HealthDeviceProfile = {
    '0200': 'SupportFeaturesList',
    '0301': 'DataExchangeSpecification',
    '0302': 'MCAP Supported Procedures',
}

__HIDProfile = {
    '0200': 'HIDDeviceReleaseNumber (Depreciated)',
    '0201': 'HIDParserVersion',
    '0202': 'HIDDeviceSubclass',
    '0203': 'HIDCountryCode',
    '0204': 'HIDVirtualCable',
    '0205': 'HIDReconnectInitiate',
    '0206': 'HIDDescriptorList',
    '0207': 'HIDLANGIDBaseList',
    '0208': 'HIDSDPDisable (Depreciated)',
    '0209': 'HIDBatteryPower',
    '020A': 'HIDRemoteWake',
    '020B': 'HIDProfileVersion',
    '020C': 'HIDSupervisionTimeout',
    '020D': 'HIDNormallyConnectable',
    '020E': 'HIDBootDevice',
    '020F': 'HIDSSRHostMaxLatency',
    '0210': 'HIDSSRHostMinTimeout',
}

__WAPBearer = {
    '0306': 'NetworkAddress',
    '0307': 'WAPGateway',
    '0308': 'HomePageURL',
    '0309': 'WAPStackType',
}

__MessageAccessProfile = {
    '0200': 'GoepLcapPsm',
    '0315': 'MASInstanceID',
    '0316': 'SupportedMessageTypes',
    '0317': 'MapSupportedFeatures',
}

__ObjectPushProfile = {
    '0200': 'GoepL2capPsm',
    '0300': 'Service Version',
    '0303': 'Supported Formats List',
}

__PANProfile = {
    '0200': 'IpSubnet',
    '030A': 'SecurityDescription',
    '030B': 'NetAccessType',
    '030C': 'MaxNetAccessrate',
    '030D': 'IPv4Subnet',
    '030E': 'IPv4Subnet',
}

__PhoneBookAccessProfile = {
    '0200': 'GoepL2capPsm',
    '0314': 'SupportedRepositories',
    '0317': 'PbapSupportedFeatures',
}

__SyncProfile = {
    '0301': 'Supported Data Stores List',
}

__MultiProfile = {
    '0200': 'MPSD Scenarios',
    '0201': 'MPMD Scenarios',
    '0202': 'Supported Profiles & Protocols',
}

ATTRIBUTES = {
    '*': __Universal,
    '110A': __SupportedFeatureOnly,
    '110B': __SupportedFeatureOnly,
    '110C': __SupportedFeatureOnly,
    '110E': __SupportedFeatureOnly,
    '110F': __SupportedFeatureOnly,
    '111B': __BIP,
    '111C': __BIP,
    '111D': __BIP,
    '1118': __BPP,
    '1119': __BPP,
    '1120': __BPP,
    '1121': __BPP,
    '1123': __BPP,
    '1000': __ServiceDiscoveryService,
    '1001': __BrowseGroupDescriptorService,
    '1200': __DeviceIdentification,
    '1111': __Fax,
    '1109': __FTP,
    '1136': __GNSSService,
    '111E': __HandsFreeProfile,
    '111F': __HandsFreeProfile,
    '1126': __HardcopyReplacementProfile,
    '1127': __HardcopyReplacementProfile,
    '1108': __HeadsetProfile,
    '1112': __HeadsetProfile,
    '1131': __HeadsetProfile,
    '1401': __HealthDeviceProfile,
    '1402': __HealthDeviceProfile,
    '1124': __HIDProfile,
    '1113': __WAPBearer,
    '1114': __WAPBearer,
    '1132': __MessageAccessProfile,
    '1133': __MessageAccessProfile,
    '1105': __ObjectPushProfile,
    '1115': __PANProfile,
    '1116': __PANProfile,
    '1117': __PANProfile,
    '112E': __PhoneBookAccessProfile,
    '112F': __PhoneBookAccessProfile,
    '1104': __SyncProfile,
    '113B': __MultiProfile,
}



from __future__ import unicode_literals

from interface import BTSimpleInterface, BTInterface
from manager import BTManager
from adapter import BTAdapter
from exceptions import BTDeviceNotSpecifiedException
import dbus.service


class BTMedia(BTSimpleInterface):
    """
    Wrapper around dbus to encapsulate org.bluez.Media
    interface.

    The media interface allows media endpoints to be
    established in accordance with the capabilities
    of a specific media service profile.

    For example, an A2DP media endpoint could be
    created allowing data from a remote device
    to be streamed to/from the sender.

    Each media endpoint is associated with a service
    object instance that implements the required
    behaviours of the endpoint.  The service object
    must be created at a given path before it
    is registered.

    See also: :py:class:`.GenericEndpoint`
    """
    def __init__(self, adapter_id=None):
        manager = BTManager()
        if (adapter_id is None):
            adapter_path = manager.default_adapter()
        else:
            adapter_path = manager.find_adapter(adapter_id)
        BTSimpleInterface.__init__(self, adapter_path, 'org.bluez.Media')

    def register_endpoint(self, path, properties):
        """
        Register a local end point to sender, the sender can
        register as many end points as it likes.

        The properties dictionary parameter may take the
        following fields:

        * UUID(str): UUID of the profile which the endpoint is
            for e.g., ``SERVICES['AudioSource']``
        * Codec(byte): Assigned mumber of codec that the
            endpoint implements. The values should match the profile
            specification which is indicated by the UUID
            e.g., ``A2DP_CODECS['SBC']``
        * Capabilities(array{byte}): Capabilities blob, it is used
            as it is so the size and byte order must match.

        See also: :py:data:`.A2DP_CODECS` and :py:data:`.SERVICES`

        If the sender disconnects the end points are
        automatically unregistered.

        :param: str path: a freely definable path name for
            registering the endpoint e.g., '/endpoint/a2dpsource'.
        :param: dict properties: a dictionary defining the
            endpoints properties which may contain.
        :return:
        :raises dbus.Exception: org.bluez.Error.InvalidArguments
        :raises dbus.Exception: org.bluez.Error.NotSupported
        """
        self._interface.RegisterEndpoint(path, properties)

    def unregister_endpoint(self, path):
        """
        Unregister sender end point.

        :param: str path: a freely definable path name previously
            used for registering via :py:meth:`register_endpoint`
        :return:
        """
        self._interface.UnregisterEndpoint(path)


class BTMediaTransport(BTInterface):
    """
    Wrapper around dbus to encapsulate the org.bluez.MediaTransport
    interface.

    A media transport instance provides a mechanism by which
    data streamed to/from a device can be sent/received via
    the L2 transport.

    This is done by `acquiring` a transport file descriptor
    object which then allows the underlying file to be
    read from or written to.

    When finished, the file descriptor should be closed using
    a normal UNIX **close()** operation before `releasing` the
    media transport back again.

    .. warning:: The media transport interface does not implement
        any CODECs and should be regarded as providing direct
        access to the underlying data transport.

    :param: str path: [Optional] Absolute object path to the
        device's file descriptor node carrying the media transport.
    :param: str dev_path: [Optional] Device's object path.  The
        file descriptor node will be provided via `fd` argument.
    :param: int fd: [Optional] File number for the file descriptor
        node associated with the device given by either `dev_path`
        or `dev_id`
    :param: str adapter_id: [Optional] Adapter identifier from
        which to look-up the `dev_id`
    :param: str dev_id: [Optional] Device identifier of the form
        e.g., '11:22:33:44:55:66'.
    :raises BTDeviceNotSpecifiedException: if the device path could
        not be determined unambiguously from the supplied parameters.

    :Properties:

    * **Device(str) [readonly]**: Device object which the transport
        is connected to.
    * **UUID(str) [readonly]**: UUID of the profile which the
        transport is for.
    * **Codec(byte) [readonly]**: Assigned mumber of codec that the
        transport support.  The value should match the profile
        specification which is indicated by the UUID.
    * **Configuration(byte) [readonly]**: Configuration blob, it is
        used as it is so the size and byte order must match.
    * **Delay(uint16) [readwrite]**: Optional. Transport delay in
            1/10 of milisecond, this property is only writeable when
            the transport was acquired by the sender.
    * **NREC(boolean) [readwrite]**: Optional. Indicates if echo
        cancelling and noise reduction functions are active in the
        transport, this property is only writeable when the transport
        was acquired by the sender.
    * **InbandRingtone(boolean) [readwrite]**: Optional. Indicates
        if the transport support sending ringtones, this property
        is only writeable when the transport was acquired by the
        sender.
    * **Routing(str) [readonly]**: Optional. Indicates where is the
        transport being routed and may be 'HCI' or 'PCM'.
    """
    def __init__(self, path, fd=None, adapter_id=None,
                 dev_path=None, dev_id=None):
        if (not path):
            fd_suffix = '/fd' + str(fd)
            if (dev_path):
                path = dev_path + fd_suffix
            elif (dev_id):
                if (adapter_id):
                    adapter = BTAdapter(adapter_id)
                else:
                    adapter = BTAdapter()
                    path = adapter.find_device(dev_id) + fd_suffix
            else:
                raise BTDeviceNotSpecifiedException
        BTInterface.__init__(self, path, 'org.bluez.MediaTransport')

    def acquire(self, access_type):
        """
        Acquire transport file descriptor and the MTU for read
        and/or write respectively.  Possible access_type:

        * "r" : Read only access
        * "w" : Write only access
        * "rw": Read and write access

        :param str access_type: as defined above.
        :return: A tuple of the form (fd, write_mtu, read_mtu)
        :rtype: tuple
        """
        return self._interface.Acquire(access_type)

    def release(self, access_type):
        """
        Releases file descriptor.

        Possible access_type:

        * "r" : Read only access
        * "w" : Write only access
        * "rw": Read and write access

        :param str access_type: as defined above.
        """
        return self._interface.Release(access_type)


class GenericEndpoint(dbus.service.Object):
    """
    Generic media endpoint service object class.

    .. note:: GenericEndpoint can't be directly instantiated.
        It should be sub-classed and provides a template for
        implementing an endpoint service object.

    :param str path: Freely definable object path for the
        media endpoint e.g., '/endpoint/a2dpsink'.
    """
    def __init__(self, path):
        bus = dbus.SystemBus()
        super(GenericEndpoint, self).__init__(bus, path)

    def get_properties(self):
        """
        Returns the properties of the endpoint.  These should
        be initialized by a suitable subclass implementation
        by setting the `properties` class attribute.

        :return properties: dictionary of endpoint's capabilties
        :rtype: dict
        """
        return self.properties

    # Service object entry points defined below here --
    # you will need to implement these in your subclass
    @dbus.service.method("org.bluez.MediaEndpoint",
                         in_signature="", out_signature="")
    def Release(self):
        """
        Called by bluez to let us know our registration
        has been released and the endpoint no longer exists

        :return:
        """
        pass

    @dbus.service.method("org.bluez.MediaEndpoint",
                         in_signature="", out_signature="")
    def ClearConfiguration(self):
        """
        Called by bluez to let us know that the audio
        streaming process has been reset, for whatever reason,
        and we should now perform clean-up.

        :return:
        """
        pass

    @dbus.service.method("org.bluez.MediaEndpoint",
                         in_signature="oay", out_signature="")
    def SetConfiguration(self, transport, config):
        """
        Provides a path to the media transport to use and
        the active configuration that was negotiated.

        :param str transport: Path to a device's file descriptor node
            which can be used to acquire the media transport e.g.,
            '/org/bluez/985/hci0/dev_00_11_67_D2_AB_EE/fd0'
        :param array{byte} config: The configuration being used for
            the media transport.

        See also: :py:class:`.BTMediaTransport`
        """
        pass

    @dbus.service.method("org.bluez.MediaEndpoint",
                         in_signature="ay", out_signature="ay")
    def SelectConfiguration(self, caps):
        """
        Initiates negotiations of the capabilities which should
        be resolved by this method.

        :param dict caps: Dictionary of device's capabilities
            for resolving capability negotiation by comparing
            to own capabilities.
        :return config: The resolved configuration
            to be used for the media transport.
        :rtype: array{byte}
        """
        pass



from __future__ import unicode_literals

from audio import BTAudio
from device import BTGenericDevice


class BTHeadset(BTAudio):
    """Wrapper around Dbus to encapsulate the BT headset entity"""

    SIGNAL_ANSWER_REQUESTED = 'AnswerRequested'
    SIGNAL_SPEAKER_GAIN_CHANGED = 'SpeakerGainChanged'
    SIGNAL_MICROPHONE_GAIN_CHANGED = 'MicrophoneGainChanged'

    def __init__(self, *args, **kwargs):
        BTGenericDevice.__init__(self, addr='org.bluez.Headset',
                                 *args, **kwargs)
        self._register_signal_name(BTHeadset.SIGNAL_ANSWER_REQUESTED)
        self._register_signal_name(BTHeadset.SIGNAL_SPEAKER_GAIN_CHANGED)
        self._register_signal_name(BTHeadset.SIGNAL_MICROPHONE_GAIN_CHANGED)

    def is_connected(self):
        return self._interface.IsConnected()

    def indicate_call(self):
        """Indicate an incoming call on the headset
        connected to the stream. Will continue to
        ring the headset about every 3 seconds."""
        return self._interface.IndicateCall()

    def cancel_call(self):
        """Cancel the incoming call indication"""
        return self._interface.CancelCall()

    def play(self):
        """Open the audio connection to the headset"""
        return self._interface.Play()

    def stop(self):
        """Close the audio connection"""
        return self._interface.Stop()


class BTHeadsetGateway(BTAudio):
    """Wrapper around Dbus to encapsulate the BT headset
    gateway entity"""

    SIGNAL_RING = 'Ring'
    SIGNAL_CALL_TERMINATED = 'CallTerminated'
    SIGNAL_CALL_STARTED = 'CallStarted'
    SIGNAL_CALL_ENDED = 'CallEnded'

    def __init__(self, *args, **kwargs):
        BTGenericDevice.__init__(self, addr='org.bluez.Headset',
                                 *args, **kwargs)
        self._register_signal_name(BTHeadsetGateway.SIGNAL_RING)
        self._register_signal_name(BTHeadsetGateway.SIGNAL_CALL_TERMINATED)
        self._register_signal_name(BTHeadsetGateway.SIGNAL_CALL_STARTED)
        self._register_signal_name(BTHeadsetGateway.SIGNAL_CALL_ENDED)

    def answer_call(self):
        """It has to called only after Ring signal received."""
        return self._interface.AnswerCall()

    def terminate_call(self):
        """Terminate call which is running or reject an incoming
        call. This has nothing with any 3-way situation incl.
        RaH. Just plain old PDH."""
        return self._interface.TerminateCall()

    def call(self, dial_number):
        """Dial a number. No number processing is done
        thus if AG would reject to dial it don't blame me"""
        return self._interface.Call(dial_number)

    def get_operator_name(self):
        """Find out the name of the currently selected network
        operator by AG."""
        return self._interface.GetOperatorName()

    def send_dtmf(self, digits):
        """Will send each digit in the 'digits' sequentially. Would
        send nothing if there is non-DTMF digit."""
        return self._interface.SendDTMF(digits)

    def get_subscriber_number(self):
        """Get the voicecall subscriber number of AG"""
        return self._interface.GetSubscriberNumber()



from __future__ import unicode_literals

from device import BTGenericDevice


class BTControl(BTGenericDevice):
    """Wrapper around Dbus to encapsulate the BT control entity"""

    SIGNAL_CONNECTED = 'Connected'
    SIGNAL_DISCONNECTED = 'Disconnected'

    def __init__(self, *args, **kwargs):
        BTGenericDevice.__init__(self, addr='org.bluez.Control',
                                 *args, **kwargs)
        self._register_signal_name(BTControl.SIGNAL_CONNECTED)
        self._register_signal_name(BTControl.SIGNAL_DISCONNECTED)

    def is_connected(self):
        return self._interface.IsConnected()

    def volume_up(self):
        """Adjust remote volume one step up"""
        self._interface.VolumeUp()

    def volume_down(self):
        """Adjust remote volume one step down"""
        self._interface.VolumeDown()



from __future__ import unicode_literals

from interface import BTInterface
from manager import BTManager


class BTAdapter(BTInterface):
    """
    Wrapper around dbus to encapsulate org.bluez.adapter interface.

    :param str adapter_path: Object path to bluetooth adapter.
        If not given, can use adapter_id instead.
    :param str adapter_id: Adapter's MAC address to look-up to find
        path e.g., '11:22:33:44:55:66'

    :Properties:

    * **Address(str) [readonly]**: The Bluetooth device address
        of the adapter.
    * **Name(str) [readonly]**: The Bluetooth system name
        (pretty hostname).
        This property is either a static system default
        or controlled by an external daemon providing
        access to the pretty hostname configuration.
    * **Alias(str) [readwrite]**: The Bluetooth friendly name.
        This value can be changed.
        In case no alias is set, it will return the system
        provided name. Setting an empty string as alias will
        convert it back to the system provided name.
        When resetting the alias with an empty string, the
        property will default back to system name.
        On a well configured system, this property never
        needs to be changed since it defaults to the system
        name and provides the pretty hostname. Only if the
        local name needs to be different from the pretty
        hostname, this property should be used as last
        resort.
    * **Class(uint32) [readonly]**: The Bluetooth class of
        device.
        This property represents the value that is either
        automatically configured by DMI/ACPI information
        or provided as static configuration.
    * **Powered(boolean) [readwrite]**: Switch an adapter on or
        off. This will also set the appropriate connectable
        state of the controller.
        The value of this property is not persistent. After
        restart or unplugging of the adapter it will reset
        back to false.
    * **Discoverable(boolean) [readwrite]**: Switch an adapter
        to discoverable or non-discoverable to either make it
        visible or hide it. This is a global setting and should
        only be used by the settings application.
        If DiscoverableTimeout is set to a non-zero
        value then the system will set this value back to
        false after the timer expired.
        In case the adapter is switched off, setting this
        value will fail.
        When changing the Powered property the new state of
        this property will be updated via a
        :py:attr:`.BTInterface.SIGNAL_PROPERTY_CHANGED`
        signal.
        For any new adapter this settings defaults to false.
    * **Pairable(boolean) [readwrite]**: Switch an adapter to
        pairable or non-pairable. This is a global setting and
        should only be used by the settings application.
    * **PairableTimeout(uint32) [readwrite]**:
        The pairable timeout in seconds. A value of zero
        means that the timeout is disabled and it will stay in
        pairable mode forever.
        The default value for pairable timeout should be
        disabled (value 0).
    * **DiscoverableTimeout(uint32) [readwrite]**:
        The discoverable timeout in seconds. A value of zero
        means that the timeout is disabled and it will stay in
        discoverable/limited mode forever.
        The default value for the discoverable timeout should
        be 180 seconds (3 minutes).
    * **Discovering(boolean) [readonly]**:
        Indicates that a device discovery procedure is active.
    * **UUIDs(array{str}) [readonly]**:
        List of 128-bit UUIDs that represents the available
        local services.
    * **Modalias(str) [readonly, optional]**:
        Local Device ID information in modalias format
        used by the kernel and udev.

    See also: :py:class:`.BTManager`
    """

    SIGNAL_DEVICE_FOUND = 'DeviceFound'
    """
    :signal DeviceFound(signal_name, user_arg, device_path): Signal
        notifying when a new device has been found
    """
    SIGNAL_DEVICE_REMOVED = 'DeviceRemoved'
    """
    :signal DeviceRemoved(signal_name, user_arg, device_path):
        Signal notifying when a device has been removed
    """
    SIGNAL_DEVICE_CREATED = 'DeviceCreated'
    """
    :signal DeviceCreated(signal_name, user_arg, device_path):
        Signal notifying when a new device is created
    """
    SIGNAL_DEVICE_DISAPPEARED = 'DeviceDisappeared'
    """
    :signal DeviceDisappeared(signal_name, user_arg, device_path):
        Signal notifying when a device is now out-of-range
    """

    def __init__(self, adapter_path=None, adapter_id=None):
        manager = BTManager()
        if (adapter_path is None):
            if (adapter_id is None):
                adapter_path = manager.default_adapter()
            else:
                adapter_path = manager.find_adapter(adapter_id)
        BTInterface.__init__(self, adapter_path, 'org.bluez.Adapter')
        self._register_signal_name(BTAdapter.SIGNAL_DEVICE_FOUND)
        self._register_signal_name(BTAdapter.SIGNAL_DEVICE_REMOVED)
        self._register_signal_name(BTAdapter.SIGNAL_DEVICE_CREATED)
        self._register_signal_name(BTAdapter.SIGNAL_DEVICE_DISAPPEARED)

    def start_discovery(self):
        """
        This method starts the device discovery session. This
        includes an inquiry procedure and remote device name
        resolving. Use :py:meth:`stop_discovery` to release the sessions
        acquired.

        This process will start emitting :py:attr:`SIGNAL_DEVICE_FOUND`
        and :py:attr:`.SIGNAL_PROPERTY_CHANGED` 'discovering' signals.

        :return:
        :raises dbus.Exception: org.bluez.Error.NotReady
        :raises dbus.Exception: org.bluez.Error.Failed
        """
        return self._interface.StartDiscovery()

    def stop_discovery(self):
        """
        This method will cancel any previous :py:meth:`start_discovery`
        transaction.

        Note that a discovery procedure is shared between all
        discovery sessions thus calling py:meth:`stop_discovery` will
        only release a single session.

        :return:
        :raises dbus.Exception: org.bluez.Error.NotReady
        :raises dbus.Exception: org.bluez.Error.Failed
        :raises dbus.Exception: org.bluez.Error.NotAuthorized
        """
        return self._interface.StopDiscovery()

    def find_device(self, dev_id):
        """
        Returns the object path of device for given address.
        The device object needs to be first created via
        :py:meth:`create_device` or
        :py:meth:`create_paired_device`

        :param str dev_id: Device MAC address to look-up e.g.,
            '11:22:33:44:55:66'
        :return: Device object path e.g.,
            '/org/bluez/985/hci0/dev_00_11_67_D2_AB_EE'
        :rtype: str
        :raises dbus.Exception: org.bluez.Error.DoesNotExist
        :raises dbus.Exception: org.bluez.Error.InvalidArguments
        """
        return self._interface.FindDevice(dev_id)

    def list_devices(self):
        """
        Returns list of device object paths.

        :return: List of device object paths
        :rtype: list
        :raises dbus.Exception: org.bluez.Error.InvalidArguments
        :raises dbus.Exception: org.bluez.Error.Failed
        :raises dbus.Exception: org.bluez.Error.OutOfMemory
        """
        return self._interface.ListDevices()

    def create_paired_device(self, dev_id, agent_path,
                             capability, cb_notify_device, cb_notify_error):
        """
        Creates a new object path for a remote device. This
        method will connect to the remote device and retrieve
        all SDP records and then initiate the pairing.

        If a previously :py:meth:`create_device` was used
        successfully, this method will only initiate the pairing.

        Compared to :py:meth:`create_device` this method will
        fail if the pairing already exists, but not if the object
        path already has been created. This allows applications
        to use :py:meth:`create_device` first and then, if needed,
        use :py:meth:`create_paired_device` to initiate pairing.

        The agent object path is assumed to reside within the
        process (D-Bus connection instance) that calls this
        method. No separate registration procedure is needed
        for it and it gets automatically released once the
        pairing operation is complete.

        :param str dev_id: New device MAC address create
            e.g., '11:22:33:44:55:66'
        :param str agent_path: Path used when creating the
            bluetooth agent e.g., '/test/agent'
        :param str capability: Pairing agent capability
            e.g., 'DisplayYesNo', etc
        :param func cb_notify_device: Callback on success.  The
            callback is called with the new device's object
            path as an argument.
        :param func cb_notify_error: Callback on error.  The
            callback is called with the error reason.
        :return:
        :raises dbus.Exception: org.bluez.Error.InvalidArguments
        :raises dbus.Exception: org.bluez.Error.Failed
        """
        return self._interface.CreatePairedDevice(dev_id,
                                                  agent_path,
                                                  capability,
                                                  reply_handler=cb_notify_device,  # noqa
                                                  error_handler=cb_notify_error)   # noqa

    def remove_device(self, dev_path):
        """
        This removes the remote device object at the given
        path. It will remove also the pairing information.

        :param str dev_path: Device object path to remove
            e.g., '/org/bluez/985/hci0/dev_00_11_67_D2_AB_EE'
        :return:
        :raises dbus.Exception: org.bluez.Error.InvalidArguments
        :raises dbus.Exception: org.bluez.Error.Failed
        """
        return self._interface.RemoveDevice(dev_path)

    def register_agent(self, path, capability):
        """
        This registers the adapter wide agent.

        The object path defines the path the of the agent
        that will be called when user input is needed.

        If an application disconnects from the bus all
        of its registered agents will be removed.

        :param str path: Freely definable path for agent e.g., '/test/agent'
        :param str capability: The capability parameter can have the values
            "DisplayOnly", "DisplayYesNo", "KeyboardOnly" and "NoInputNoOutput"
            which reflects the input and output capabilities of the agent.
            If an empty string is used it will fallback to "DisplayYesNo".
        :return:
        :raises dbus.Exception: org.bluez.Error.InvalidArguments
        :raises dbus.Exception: org.bluez.Error.AlreadyExists
        """
        return self._interface.RegisterAgent(path, capability)

    def unregister_agent(self, path):
        """
        This unregisters the agent that has been previously
        registered. The object path parameter must match the
        same value that has been used on registration.

        :param str path: Previously defined path for agent
            e.g., '/test/agent'
        :return:
        :raises dbus.Exception: org.bluez.Error.DoesNotExist
        """
        return self._interface.UnregisterAgent(path)



from __future__ import unicode_literals

from interface import BTInterface
from adapter import BTAdapter
from exceptions import BTDeviceNotSpecifiedException


class BTGenericDevice(BTInterface):
    """
    Generic BT device which has its own interface bus address but is
    associated with a BT adapter.

    :param str addr: Interface address e.g., 'org.bluez.Device'
    :param str dev_path: Optional fully qualified device object path
        e.g., '/org/bluez/985/hci0/dev_00_11_67_D2_AB_EE'
    :param str adapter_path: Optional fully qualified adapter object
        path if using dev_id e.g., '/org/bluez/985/hci0'
    :param str adapter_id: Optional adapter device ID if using dev_id
        e.g. 'hci0' or '11:22:33:44:55:66'
    :param str dev_id: Optional device ID to look-up object path
        e.g. '11:22:33:44:55:66'
    :raises BTDeviceNotSpecifiedException:
        if the device address was not specified unambiguously.

    .. note:: This class should always be sub-classed with a concrete
        implementation of a bluez interface.
    """
    def __init__(self, addr, dev_path=None, adapter_path=None,
                 adapter_id=None, dev_id=None):
        if (dev_path):
            path = dev_path
        elif (dev_id):
            if (adapter_id or adapter_path):
                adapter = BTAdapter(adapter_path=adapter_path,
                                    adapter_id=adapter_id)
            else:
                adapter = BTAdapter()
            path = adapter.find_device(dev_id)
        else:
            raise BTDeviceNotSpecifiedException
        BTInterface.__init__(self, path, addr)


class BTDevice(BTGenericDevice):
    """
    Wrapper around dbus to encapsulate the org.bluez.Device interface.
    Refer to :py:class:`BTGenericDevice` for init arguments.

    :Properties:

    * **Address(str) [readonly]**: The Bluetooth device address of the
        remote device.
    * **Name(str) [readonly]**: The Bluetooth remote name. This value can
        not be changed. Use the Alias property instead.
    * **Icon(str) [readonly]**: Proposed icon name according to the
        freedesktop.org icon naming specification.
    * **Class(str) [readonly]**: The Bluetooth class of device of the remote
        device.
    * **UUIDs(list{str}) [readonly]**: List of 128-bit UUIDs that represents
        the available remote services.
    * **Paired(boolean) [readonly]**: Indicates if the remote device is paired.
    * **Connected(boolean) [readonly]**: Indicates if the remote device is
        connected.
    * **Trusted(boolean) [readwrite]**: Indicates if the remote device is
        trusted.
    * **Alias(str) [readwrite]**: The name alias for the remote device. The
        alias can be used to have a different friendly name for the remote
        device. In case no alias is set, it will return the remote device name.
        Setting an empty string as alias will convert it back to the remote
        device name. When reseting the alias with an empty string, the emitted
        :py:attr`.SIGNAL_PROPERTY_CHANGED` signal will show the remote
        name again.
    * **Nodes(list{str}) [readonly]**: List of device node object paths.
    * **Adapter(str) [readonly]**: The object path of the adpater the device
        belongs to.
    * **LegacyPairing(boolean) [readonly]**: Set to true if the device only
        supports the pre-2.1 pairing mechanism. This property is useful in the
        :py:attr`.SIGNAL_DEVICE_FOUND` signal to anticipate whether legacy
        or simple pairing will occur. Note that this property can exhibit
        false-positives in the case of Bluetooth 2.1 (or newer) devices that
        have disabled Extended Inquiry Response support.
    * **Blocked(boolean) [readonly]**: Set to true if the device is blocked.
    * **Product(uint16) [readonly]**: Product code identifier.
    * **Vendor(uint16) [readonly]**: Product vendor identifier.
    * **Services(list{str}) [readonly]**: TBD.
    """

    SIGNAL_DISCONNECT_REQUESTED = 'DisconnectRequested'
    """
    :signal DisconnectRequested(signal_name, user_arg): This signal will be
        sent when a low level disconnection to a remote device has been
        requested. The actual disconnection will happen 2 seconds later.
    """
    SIGNAL_NODE_CREATED = 'NodeCreated'
    """
    :signal NodeCreated(signal_name, user_arg, node_path): Signal notifying
        when a new device node has been created.
    """
    SIGNAL_NODE_REMOVED = 'NodeRemoved'
    """
    :signal NodeRemoved(signal_name, user_arg, node_path): Signal notifying
        when a device node has been removed.
    """

    def __init__(self, *args, **kwargs):
        BTGenericDevice.__init__(self, addr='org.bluez.Device',
                                 *args, **kwargs)
        self._register_signal_name(BTDevice.SIGNAL_DISCONNECT_REQUESTED)
        self._register_signal_name(BTDevice.SIGNAL_NODE_CREATED)
        self._register_signal_name(BTDevice.SIGNAL_NODE_REMOVED)

    def discover_services(self, pattern=''):
        """
        This method starts the service discovery to retrieve
        remote service records.

        Refer to :py:class:`.BTDiscoveryInfo` to decode each XML service
        record.

        :param str pattern: can be used to specify specific UUIDs.
            An empty string will look for the public browse group.
        :return: a dictionary with the record handles as keys and
            the service record in XML format as values. The key is
            uint32 and the value a string for this dictionary.
        :rtype: dict
        :raises dbus.Exception: org.bluez.Error.NotReady
        :raises dbus.Exception: org.bluez.Error.Failed
        :raises dbus.Exception: org.bluez.Error.InProgress
        """
        return self._interface.DiscoverServices(pattern)

    def cancel_discovery(self):
        """
        This method will cancel any previous :py:meth:`discover_services`
        transaction.

        :return:
        :raises dbus.Exception: org.bluez.Error.NotReady
        :raises dbus.Exception: org.bluez.Error.Failed
        :raises dbus.Exception: org.bluez.Error.NotAuthorized
        """
        return self._interface.CancelDiscovery()

    def disconnect(self):
        """
        This method disconnects a specific remote device by
        terminating the low-level ACL connection. The use of
        this method should be restricted to administrator
        use.

        A :py:attr:`.SIGNAL_DISCONNECT_REQUESTED` signal will be
        sent and the actual disconnection will only happen 2
        seconds later.
        This enables upper-level applications to terminate
        their connections gracefully before the ACL connection
        is terminated.

        :return:
        :raises dbus.Exception: org.bluez.Error.NotConnected
        """
        return self._interface.Disconnect()



from __future__ import unicode_literals

from uuid import BTUUID16

__SDP = BTUUID16('0001', 'SDP', 'Bluetooth Core Specification')  # noqa
__UDP = BTUUID16('0002', 'UDP', '[NO USE BY PROFILES]')  # noqa
__RFCOMM = BTUUID16('0003', 'RFCOMM', 'RFCOMM with TS 07.10')  # noqa
__TCP = BTUUID16('0004', 'TCP', '[NO USE BY PROFILES]')  # noqa
__TCS_BIN = BTUUID16('0005', 'TCS_BIN', 'Telephony Control Specification / TCS Binary [DEPRECATED]')  # noqa
__TCS_AT = BTUUID16('0006', 'TCS_AT', '[NO USE BY PROFILES]')  # noqa
__ATT = BTUUID16('0007', 'ATT', 'Attribute Protocol')  # noqa
__OBEX = BTUUID16('0008', 'OBEX', 'IrDA Interoperability')  # noqa
__IP = BTUUID16('0009', 'IP', '[NO USE BY PROFILES]')  # noqa
__FTP = BTUUID16('000A', 'FTP', '[NO USE BY PROFILES]')  # noqa
__HTTP = BTUUID16('000C', 'HTTP', '[NO USE BY PROFILES]')  # noqa
__WSP = BTUUID16('000E', 'WSP', '[NO USE BY PROFILES]')  # noqa
__BNEP = BTUUID16('000F', 'BNEP', 'Bluetooth Network Encapsulation Protocol (BNEP)')  # noqa
__UPNP = BTUUID16('0010', 'UPNP', 'Extended Service Discovery Profile (ESDP) [DEPRECATED]')  # noqa
__HIDP = BTUUID16('0011', 'HIDP', 'Human Interface Device Profile (HID)')  # noqa
__HardcopyControlChannel = BTUUID16('0012', 'HardcopyControlChannel', 'Hardcopy Cable Replacement Profile (HCRP)')  # noqa
__HardcopyDataChannel = BTUUID16('0014', 'HardcopyDataChannel', 'See Hardcopy Cable Replacement Profile (HCRP)')  # noqa
__HardcopyNotification = BTUUID16('0016', 'HardcopyNotification', 'Hardcopy Cable Replacement Profile (HCRP)')  # noqa
__AVCTP = BTUUID16('0017', 'AVCTP', 'Audio/Video Control Transport Protocol (AVCTP)')  # noqa
__AVDTP = BTUUID16('0019', 'AVDTP', 'Audio/Video Distribution Transport Protocol (AVDTP)')  # noqa
__CMTP = BTUUID16('001B', 'CMTP', 'Common ISDN Access Profile (CIP) [DEPRECATED]')  # noqa
__MCAPControlChannel = BTUUID16('001E', 'MCAPControlChannel', 'Multi-Channel Adaptation Protocol (MCAP)')  # noqa
__MCAPDataChannel = BTUUID16('001F', 'MCAPDataChannel', 'Multi-Channel Adaptation Protocol (MCAP)')  # noqa
__L2CAP = BTUUID16('0100', 'L2CAP', 'Bluetooth Core Specification')  # noqa

__ServiceDiscoveryServerServiceClassID = BTUUID16('1000', 'ServiceDiscoveryServerServiceClassID', 'Bluetooth Core Specification')  # noqa
__BrowseGroupDescriptorServiceClassID = BTUUID16('1001', 'BrowseGroupDescriptorServiceClassID', 'Bluetooth Core Specification')  # noqa
__PublicBrowseRoot = BTUUID16('1002', 'PublicBrowseRoot', 'Bluetooth Core Specification')  # noqa
__SerialPort = BTUUID16('1101', 'SerialPort', 'Serial Port Profile (SPP) NOTE: __The example SDP record in SPP v1.0 does not include a BluetoothProfileDescriptorList attribute, but some implementations may also use this UUID for the Profile Identifier.')  # noqa
__LANAccessUsingPPP = BTUUID16('1102', 'LANAccessUsingPPP', 'LAN Access Profile [DEPRECATED] NOTE: Used as both Service Class Identifier and Profile Identifier.')  # noqa
__DialupNetworking = BTUUID16('1103', 'DialupNetworking', 'Dial-up Networking Profile (DUN) NOTE: Used as both Service Class Identifier and Profile Identifier.')  # noqa
__IrMCSync = BTUUID16('1104', 'IrMCSync', 'Synchronization Profile (SYNC) NOTE: Used as both Service Class Identifier and Profile Identifier.')  # noqa
__OBEXObjectPush = BTUUID16('1105', 'OBEXObjectPush', 'Object Push Profile (OPP) NOTE: Used as both Service Class Identifier and Profile.')  # noqa
__OBEXFileTransfer = BTUUID16('1106', 'OBEXFileTransfer', 'File Transfer Profile (FTP) NOTE: Used as both Service Class Identifier and Profile Identifier.')  # noqa
__IrMCSyncCommand = BTUUID16('1107', 'IrMCSyncCommand', 'Synchronization Profile (SYNC)')  # noqa
__Headset = BTUUID16('1108', 'Headset', 'Headset Profile (HSP) NOTE: Used as both Service Class Identifier and Profile Identifier.')  # noqa
__CordlessTelephony = BTUUID16('1109', 'CordlessTelephony', 'Cordless Telephony Profile (CTP) NOTE: Used as both Service Class Identifier and Profile Identifier. [DEPRECATED]')  # noqa
__AudioSource = BTUUID16('110A', 'AudioSource', 'Advanced Audio Distribution Profile (A2DP)')  # noqa
__AudioSink = BTUUID16('110B', 'AudioSink', 'Advanced Audio Distribution Profile (A2DP)')  # noqa
__AVRemoteControlTarget = BTUUID16('110C', 'AVRemoteControlTarget', 'Audio/Video Remote Control Profile (AVRCP)')  # noqa
__AdvancedAudioDistribution = BTUUID16('110D', 'AdvancedAudioDistribution', 'Advanced Audio Distribution Profile (A2DP)')  # noqa
__AVRemoteControl = BTUUID16('110E', 'AVRemoteControl', 'Audio/Video Remote Control Profile (AVRCP) NOTE: Used as both Service Class Identifier and Profile Identifier.')  # noqa
__AVRemoteControlController = BTUUID16('110F', 'AVRemoteControlController', 'Audio/Video Remote Control Profile (AVRCP) NOTE: __The AVRCP specification v1.3 and later require that 0x110E also be included in the ServiceClassIDList before 0x110F for backwards compatibility.')  # noqa
__Intercom = BTUUID16('1110', 'Intercom', 'Intercom Profile (ICP) NOTE: Used as both Service Class Identifier and Profile Identifier. [DEPRECATED]')  # noqa
__Fax = BTUUID16('1111', 'Fax', 'Fax Profile (FAX) NOTE: Used as both Service Class Identifier and Profile Identifier. [DEPRECATED]')  # noqa
__HeadsetAudioGateway = BTUUID16('1112', 'HeadsetAudioGateway', 'Headset Profile (HSP)')  # noqa
__WAP = BTUUID16('1113', 'WAP', 'Interoperability Requirements for Bluetooth technology as a WAP, Bluetooth SIG [DEPRECATED]')  # noqa
__WAPCLIENT = BTUUID16('1114', 'WAPCLIENT', 'Interoperability Requirements for Bluetooth technology as a WAP, Bluetooth SIG [DEPRECATED]')  # noqa
__PANU = BTUUID16('1115', 'PANU', 'Personal Area Networking Profile (PAN) NOTE: Used as both Service Class Identifier and Profile Identifier for PANU role.')  # noqa
__NAP = BTUUID16('1116', 'NAP', 'Personal Area Networking Profile (PAN) NOTE: Used as both Service Class Identifier and Profile Identifier for NAP role.')  # noqa
__GN = BTUUID16('1117', 'GN', 'Personal Area Networking Profile (PAN) NOTE: Used as both Service Class Identifier and Profile Identifier for GN role.')  # noqa
__DirectPrinting = BTUUID16('1118', 'DirectPrinting', 'Basic Printing Profile (BPP)')  # noqa
__ReferencePrinting = BTUUID16('1119', 'ReferencePrinting', 'See Basic Printing Profile (BPP)')  # noqa
__BasicImagingProfile = BTUUID16('111A', 'BasicImagingProfile', 'Basic Imaging Profile (BIP)')  # noqa
__ImagingResponder = BTUUID16('111B', 'ImagingResponder', 'Basic Imaging Profile (BIP)')  # noqa
__ImagingAutomaticArchive = BTUUID16('111C', 'ImagingAutomaticArchive', 'Basic Imaging Profile (BIP)')  # noqa
__ImagingReferencedObjects = BTUUID16('111D', 'ImagingReferencedObjects', 'Basic Imaging Profile (BIP)')  # noqa
__Handsfree = BTUUID16('111E', 'Handsfree', 'Hands-Free Profile (HFP) NOTE: Used as both Service Class Identifier and Profile Identifier.')  # noqa
__HandsfreeAudioGateway = BTUUID16('111F', 'HandsfreeAudioGateway', 'Hands-free Profile (HFP)')  # noqa
__DirectPrintingReferenceObjectsService = BTUUID16('1120', 'DirectPrintingReferenceObjectsService', 'Basic Printing Profile (BPP)')  # noqa
__ReflectedUI = BTUUID16('1121', 'ReflectedUI', 'Basic Printing Profile (BPP)')  # noqa
__BasicPrinting = BTUUID16('1122', 'BasicPrinting', 'Basic Printing Profile (BPP)')  # noqa
__PrintingStatus = BTUUID16('1123', 'PrintingStatus', 'Basic Printing Profile (BPP)')  # noqa
__HumanInterfaceDeviceService = BTUUID16('1124', 'HumanInterfaceDeviceService', 'Human Interface Device (HID) NOTE: Used as both Service Class Identifier and Profile Identifier.')  # noqa
__HardcopyCableReplacement = BTUUID16('1125', 'HardcopyCableReplacement', 'Hardcopy Cable Replacement Profile (HCRP)')  # noqa
__HCRPrint = BTUUID16('1126', 'HCRPrint', 'Hardcopy Cable Replacement Profile (HCRP)')  # noqa
__HCRScan = BTUUID16('1127', 'HCRScan', 'Hardcopy Cable Replacement Profile (HCRP)')  # noqa
__CommonISDNAccess = BTUUID16('1128', 'CommonISDNAccess', 'Common ISDN Access Profile (CIP) NOTE: Used as both Service Class Identifier and Profile Identifier. [DEPRECATED]')  # noqa
__VideoConfGateway = BTUUID16('1129', 'VideoConfGateway', 'Video Conference Gateway')  # noqa
__UDIMtClass = BTUUID16('112A', 'UDIMtClass', '[UDI]')  # noqa
__UDITaClass = BTUUID16('112B', 'UDITaClass', '[UDI]')  # noqa
__AudioVideo = BTUUID16('112C', 'AudioVideo', 'Video Conferencing Profile (VCP), Bluetooth SIG')  # noqa
__SIMAccess = BTUUID16('112D', 'SIMAccess', 'SIM Access Profile (SAP) NOTE: Used as both Service Class Identifier and Profile Identifier.')  # noqa
__PhonebookAccessPCE = BTUUID16('112E', 'PhonebookAccessPCE', 'Phonebook Access Profile (PBAP)')  # noqa
__PhonebookAccessPSE = BTUUID16('112F', 'PhonebookAccessPSE', 'Phonebook Access Profile (PBAP)')  # noqa
__PhonebookAccess = BTUUID16('1130', 'PhonebookAccess', 'Phonebook Access Profile (PBAP)')  # noqa
__HeadsetHS = BTUUID16('1131', 'HeadsetHS', 'Headset Profile (HSP) NOTE: See erratum #3507. 0x1108 and 0x1203 should also be included in the ServiceClassIDList before 0x1131 for backwards compatibility.')  # noqa
__MessageAccessServer = BTUUID16('1132', 'MessageAccessServer', 'Message Access Profile (MAP)')  # noqa
__MessageNotificationServer = BTUUID16('1133', 'MessageNotificationServer', 'Message Access Profile (MAP)')  # noqa
__MessageAccessProfile = BTUUID16('1134', 'MessageAccessProfile', 'Message Access Profile (MAP)')  # noqa
__GNSS = BTUUID16('1135', 'GNSS', 'Global Navigation Satellite System Profile (GNSS)')  # noqa
__GNSSServer = BTUUID16('1136', 'GNSSServer', 'Global Navigation Satellite System Profile (GNSS)')  # noqa
__Display3D = BTUUID16('1137', 'Display3D', '3D Synchronization Profile (3DSP)')  # noqa
__Glasses3D = BTUUID16('1138', 'Glasses3D', '3D Synchronization Profile (3DSP)')  # noqa
__Synchronization3D = BTUUID16('1139', 'Synchronization3D', '3D Synchronization Profile (3DSP)')  # noqa
__MPSProfile = BTUUID16('113A', 'MPSProfile', 'Multi-Profile Specification (MPS)')  # noqa
__MPSSC = BTUUID16('113B', 'MPSSC', 'Multi-Profile Specification (MPS)')  # noqa
__PnPInformation = BTUUID16('1200', 'PnPInformation', 'Device Identification (DID) NOTE: Used as both Service Class Identifier and Profile Identifier.')  # noqa
__GenericNetworking = BTUUID16('1201', 'GenericNetworking', 'N/A')  # noqa
__GenericFileTransfer = BTUUID16('1202', 'GenericFileTransfer', 'N/A')  # noqa
__GenericAudio = BTUUID16('1203', 'GenericAudio', 'N/A')  # noqa
__GenericTelephony = BTUUID16('1204', 'GenericTelephony', 'N/A')  # noqa
__UPNPService = BTUUID16('1205', 'UPNPService', 'Enhanced Service Discovery Profile (ESDP) [DEPRECATED]')  # noqa
__UPNPIPService = BTUUID16('1206', 'UPNPIPService', 'Enhanced Service Discovery Profile (ESDP) [DEPRECATED]')  # noqa
__ESDPUPNPIPPAN = BTUUID16('1300', 'ESDPUPNPIPPAN', 'Enhanced Service Discovery Profile (ESDP) [DEPRECATED]')  # noqa
__ESDPUPNPIPLAP = BTUUID16('1301', 'ESDPUPNPIPLAP', 'Enhanced Service Discovery Profile (ESDP)[DEPRECATED]')  # noqa
__ESDPUPNPL2CAP = BTUUID16('1302', 'ESDPUPNPL2CAP', 'Enhanced Service Discovery Profile (ESDP)[DEPRECATED]')  # noqa
__VideoSource = BTUUID16('1303', 'VideoSource', 'Video Distribution Profile (VDP)')  # noqa
__VideoSink = BTUUID16('1304', 'VideoSink', 'Video Distribution Profile (VDP)')  # noqa
__VideoDistribution = BTUUID16('1305', 'VideoDistribution', 'Video Distribution Profile (VDP)')  # noqa
__HDP = BTUUID16('1400', 'HDP', 'Health Device Profile')  # noqa
__HDPSource = BTUUID16('1401', 'HDPSource', 'Health Device Profile (HDP)')  # noqa
__HDPSink = BTUUID16('1402', 'HDPSink', 'Health Device Profile (HDP)')  # noqa

SERVICES = {
    __SDP.uuid16: __SDP,
    __SDP.name: __SDP,
    __UDP.uuid16: __UDP,
    __UDP.name: __UDP,
    __RFCOMM.uuid16: __RFCOMM,
    __RFCOMM.name: __RFCOMM,
    __TCP.uuid16: __TCP,
    __TCP.name: __TCP,
    __TCS_BIN.uuid16: __TCS_BIN,
    __TCS_BIN.name: __TCS_BIN,
    __TCS_AT.uuid16: __TCS_AT,
    __TCS_AT.name: __TCS_AT,
    __ATT.uuid16: __ATT,
    __ATT.name: __ATT,
    __OBEX.uuid16: __OBEX,
    __OBEX.name: __OBEX,
    __IP.uuid16: __IP,
    __IP.name: __IP,
    __FTP.uuid16: __FTP,
    __FTP.name: __FTP,
    __HTTP.uuid16: __HTTP,
    __HTTP.name: __HTTP,
    __WSP.uuid16: __WSP,
    __WSP.name: __WSP,
    __BNEP.uuid16: __BNEP,
    __BNEP.name: __BNEP,
    __UPNP.uuid16: __UPNP,
    __UPNP.name: __UPNP,
    __HIDP.uuid16: __HIDP,
    __HIDP.name: __HIDP,
    __HardcopyControlChannel.uuid16: __HardcopyControlChannel,
    __HardcopyControlChannel.name: __HardcopyControlChannel,
    __HardcopyDataChannel.uuid16: __HardcopyDataChannel,
    __HardcopyDataChannel.name: __HardcopyDataChannel,
    __HardcopyNotification.uuid16: __HardcopyNotification,
    __HardcopyNotification.name: __HardcopyNotification,
    __AVCTP.uuid16: __AVCTP,
    __AVCTP.name: __AVCTP,
    __AVDTP.uuid16: __AVDTP,
    __AVDTP.name: __AVDTP,
    __CMTP.uuid16: __CMTP,
    __CMTP.name: __CMTP,
    __MCAPControlChannel.uuid16: __MCAPControlChannel,
    __MCAPControlChannel.name: __MCAPControlChannel,
    __MCAPDataChannel.uuid16: __MCAPDataChannel,
    __MCAPDataChannel.name: __MCAPDataChannel,
    __L2CAP.uuid16: __L2CAP,
    __L2CAP.name: __L2CAP,
    __ServiceDiscoveryServerServiceClassID.uuid16: __ServiceDiscoveryServerServiceClassID,  # noqa
    __ServiceDiscoveryServerServiceClassID.name: __ServiceDiscoveryServerServiceClassID,  # noqa
    __BrowseGroupDescriptorServiceClassID.uuid16: __BrowseGroupDescriptorServiceClassID,  # noqa
    __BrowseGroupDescriptorServiceClassID.name: __BrowseGroupDescriptorServiceClassID,  # noqa
    __SerialPort.uuid16: __SerialPort,
    __SerialPort.name: __SerialPort,
    __LANAccessUsingPPP.uuid16: __LANAccessUsingPPP,
    __LANAccessUsingPPP.name: __LANAccessUsingPPP,
    __DialupNetworking.uuid16: __DialupNetworking,
    __DialupNetworking.name: __DialupNetworking,
    __IrMCSync.uuid16: __IrMCSync,
    __IrMCSync.name: __IrMCSync,
    __OBEXObjectPush.uuid16: __OBEXObjectPush,
    __OBEXObjectPush.name: __OBEXObjectPush,
    __OBEXFileTransfer.uuid16: __OBEXFileTransfer,
    __OBEXFileTransfer.name: __OBEXFileTransfer,
    __IrMCSyncCommand.uuid16: __IrMCSyncCommand,
    __IrMCSyncCommand.name: __IrMCSyncCommand,
    __Headset.uuid16: __Headset,
    __Headset.name: __Headset,
    __CordlessTelephony.uuid16: __CordlessTelephony,
    __CordlessTelephony.name: __CordlessTelephony,
    __AudioSource.uuid16: __AudioSource,
    __AudioSource.name: __AudioSource,
    __AudioSink.uuid16: __AudioSink,
    __AudioSink.name: __AudioSink,
    __AVRemoteControlTarget.uuid16: __AVRemoteControlTarget,
    __AVRemoteControlTarget.name: __AVRemoteControlTarget,
    __AdvancedAudioDistribution.uuid16: __AdvancedAudioDistribution,
    __AdvancedAudioDistribution.name: __AdvancedAudioDistribution,
    __AVRemoteControl.uuid16: __AVRemoteControl,
    __AVRemoteControl.name: __AVRemoteControl,
    __AVRemoteControlController.uuid16: __AVRemoteControlController,
    __AVRemoteControlController.name: __AVRemoteControlController,
    __Intercom.uuid16: __Intercom,
    __Intercom.name: __Intercom,
    __Fax.uuid16: __Fax,
    __Fax.name: __Fax,
    __HeadsetAudioGateway.uuid16: __HeadsetAudioGateway,
    __HeadsetAudioGateway.name: __HeadsetAudioGateway,
    __WAP.uuid16: __WAP,
    __WAP.name: __WAP,
    __WAPCLIENT.uuid16: __WAPCLIENT,
    __WAPCLIENT.name: __WAPCLIENT,
    __PANU.uuid16: __PANU,
    __PANU.name: __PANU,
    __NAP.uuid16: __NAP,
    __NAP.name: __NAP,
    __GN.uuid16: __GN,
    __GN.name: __GN,
    __DirectPrinting.uuid16: __DirectPrinting,
    __DirectPrinting.name: __DirectPrinting,
    __ReferencePrinting.uuid16: __ReferencePrinting,
    __ReferencePrinting.name: __ReferencePrinting,
    __BasicImagingProfile.uuid16: __BasicImagingProfile,
    __BasicImagingProfile.name: __BasicImagingProfile,
    __ImagingResponder.uuid16: __ImagingResponder,
    __ImagingResponder.name: __ImagingResponder,
    __ImagingAutomaticArchive.uuid16: __ImagingAutomaticArchive,
    __ImagingAutomaticArchive.name: __ImagingAutomaticArchive,
    __ImagingReferencedObjects.uuid16: __ImagingReferencedObjects,
    __ImagingReferencedObjects.name: __ImagingReferencedObjects,
    __Handsfree.uuid16: __Handsfree,
    __Handsfree.name: __Handsfree,
    __HandsfreeAudioGateway.uuid16: __HandsfreeAudioGateway,
    __HandsfreeAudioGateway.name: __HandsfreeAudioGateway,
    __DirectPrintingReferenceObjectsService.uuid16: __DirectPrintingReferenceObjectsService,  # noqa
    __DirectPrintingReferenceObjectsService.name: __DirectPrintingReferenceObjectsService,  # noqa
    __ReflectedUI.uuid16: __ReflectedUI,
    __ReflectedUI.name: __ReflectedUI,
    __BasicPrinting.uuid16: __BasicPrinting,
    __BasicPrinting.name: __BasicPrinting,
    __PrintingStatus.uuid16: __PrintingStatus,
    __PrintingStatus.name: __PrintingStatus,
    __HumanInterfaceDeviceService.uuid16: __HumanInterfaceDeviceService,
    __HumanInterfaceDeviceService.name: __HumanInterfaceDeviceService,
    __HardcopyCableReplacement.uuid16: __HardcopyCableReplacement,
    __HardcopyCableReplacement.name: __HardcopyCableReplacement,
    __HCRPrint.uuid16: __HCRPrint,
    __HCRPrint.name: __HCRPrint,
    __HCRScan.uuid16: __HCRScan,
    __HCRScan.name: __HCRScan,
    __CommonISDNAccess.uuid16: __CommonISDNAccess,
    __CommonISDNAccess.name: __CommonISDNAccess,
    __SIMAccess.uuid16: __SIMAccess,
    __SIMAccess.name: __SIMAccess,
    __PhonebookAccessPCE.uuid16: __PhonebookAccessPCE,
    __PhonebookAccessPCE.name: __PhonebookAccessPCE,
    __PhonebookAccessPSE.uuid16: __PhonebookAccessPSE,
    __PhonebookAccessPSE.name: __PhonebookAccessPSE,
    __PhonebookAccess.uuid16: __PhonebookAccess,
    __PhonebookAccess.name: __PhonebookAccess,
    __HeadsetHS.uuid16: __HeadsetHS,
    __HeadsetHS.name: __HeadsetHS,
    __MessageAccessServer.uuid16: __MessageAccessServer,
    __MessageAccessServer.name: __MessageAccessServer,
    __MessageNotificationServer.uuid16: __MessageNotificationServer,
    __MessageNotificationServer.name: __MessageNotificationServer,
    __MessageAccessProfile.uuid16: __MessageAccessProfile,
    __MessageAccessProfile.name: __MessageAccessProfile,
    __GNSS.uuid16: __GNSS,
    __GNSS.name: __GNSS,
    __GNSSServer.uuid16: __GNSSServer,
    __GNSSServer.name: __GNSSServer,
    __Display3D.uuid16: __Display3D,
    __Display3D.name: __Display3D,
    __Glasses3D.uuid16: __Glasses3D,
    __Glasses3D.name: __Glasses3D,
    __Synchronization3D.uuid16: __Synchronization3D,
    __Synchronization3D.name: __Synchronization3D,
    __MPSProfile.uuid16: __MPSProfile,
    __MPSProfile.name: __MPSProfile,
    __MPSSC.uuid16: __MPSSC,
    __MPSSC.name: __MPSSC,
    __PnPInformation.uuid16: __PnPInformation,
    __PnPInformation.name: __PnPInformation,
    __GenericNetworking.uuid16: __GenericNetworking,
    __GenericNetworking.name: __GenericNetworking,
    __GenericFileTransfer.uuid16: __GenericFileTransfer,
    __GenericFileTransfer.name: __GenericFileTransfer,
    __GenericAudio.uuid16: __GenericAudio,
    __GenericAudio.name: __GenericAudio,
    __GenericTelephony.uuid16: __GenericTelephony,
    __GenericTelephony.name: __GenericTelephony,
    __UPNPService.uuid16: __UPNPService,
    __UPNPService.name: __UPNPService,
    __UPNPIPService.uuid16: __UPNPIPService,
    __UPNPIPService.name: __UPNPIPService,
    __ESDPUPNPIPPAN.uuid16: __ESDPUPNPIPPAN,
    __ESDPUPNPIPPAN.name: __ESDPUPNPIPPAN,
    __ESDPUPNPIPLAP.uuid16: __ESDPUPNPIPLAP,
    __ESDPUPNPIPLAP.name: __ESDPUPNPIPLAP,
    __ESDPUPNPL2CAP.uuid16: __ESDPUPNPL2CAP,
    __ESDPUPNPL2CAP.name: __ESDPUPNPL2CAP,
    __VideoSource.uuid16: __VideoSource,
    __VideoSource.name: __VideoSource,
    __VideoSink.uuid16: __VideoSink,
    __VideoSink.name: __VideoSink,
    __VideoDistribution.uuid16: __VideoDistribution,
    __VideoDistribution.name: __VideoDistribution,
    __HDP.uuid16: __HDP,
    __HDP.name: __HDP,
    __HDPSource.uuid16: __HDPSource,
    __HDPSource.name: __HDPSource,
    __HDPSink.uuid16: __HDPSink,
    __HDPSink.name: __HDPSink,
}
"""
:data dict SERVICES: A dictionary of service UUIDs which allows all
    bluetooth standard-based services to be keyed by either
    their short-form name or their UUID16 value.

    Example:

    ``SERVICES['AudioSource']`` shall return a :py:class:`.BTUUID16`
        denoting the A2DP audio source profile UUID.
    ``SERVICES['110A']`` shall return the same :py:class:`.BTUUID16`
        denoting the A2DP audio source profile UUID.
"""



from __future__ import unicode_literals
from xml.etree.ElementTree import XML
import pprint

from serviceuuids import SERVICES
from attributes import ATTRIBUTES


class BTDiscoveryInfo:
    """
    Parser for XML discovery service record obtained from the
    bluetooth service discovery procedure.

    The XML parser recursively and iteratively translates all known
    service UUIDs and attribute UUIDs to their human readable
    form thus allowing a BTDiscoveryInfo instance to be printed
    e.g., for debugging purposes.

    :param str text: An XML service record return as part of
        the device service discovery procedure.

    .. note:: A dictionary of XML service records is obtained
        by executing the :py:meth:`.BTDevice.discover_services`
        method.
    """
    def __init__(self, text):
        tree = XML(text)
        rec = tree.iter().next()
        if (rec is not None):
            self._uuid = None
            self.__dict__ = self._parse_element(rec)

    def _parse_element(self, elem):
        if (elem.tag == 'record'):
            return {self._parse_element(k):
                    self._parse_element(list(k)[0])
                    for k in elem.findall('attribute')}
        elif (elem.tag == 'sequence'):
            return [self._parse_element(k) for k in list(elem)]
        elif (elem.tag == 'attribute'):
            attrib_id = elem.attrib['id'][2:].upper()  # Remove leading '0x'
            if (attrib_id in ATTRIBUTES['*']):
                return ATTRIBUTES['*'][attrib_id]
            elif (ATTRIBUTES.get(self._uuid)):
                return ATTRIBUTES[self._uuid].get(attrib_id, attrib_id)
            else:
                return attrib_id
        elif (elem.tag == 'uuid'):
            self._uuid = \
                elem.attrib['value'][2:].upper()  # Remove leading '0x'
            return {'uuid': SERVICES.get(self._uuid, self._uuid)}
        elif ('value' in elem.attrib):
            return elem.attrib['value']

    def __repr__(self):
        return pprint.pformat(self.__dict__)
